import io  # type: ignore
import torch
from contextlib import redirect_stdout
import matplotlib.pyplot as plt
from speckcn2.postprocess import tags_distribution


def test_tags_distribution():
    # Create a dummy dataset
    dataset = [(torch.rand(64 * 64), torch.rand(8)) for _ in range(32)]
    test_tags = torch.rand((8, 32))
    device = torch.device('cpu')

    conf = {
        'model': {
            'name': 'test_model',
        },
        'speckle': {
            'datadirectory': 'speckcn2/assets/test'
        }
    }

    # Temporarily redirect stdout to a string buffer
    with io.StringIO() as buf, redirect_stdout(buf):
        # Call the function
        tags_distribution(conf, dataset, test_tags, device, rescale=False)

        # Now we can check if the print statements in your function are as expected
        assert 'Data shape:' in buf.getvalue()
        assert 'Prediction shape:' in buf.getvalue()
        assert 'Train mean:' in buf.getvalue()
        assert 'Train std:' in buf.getvalue()
        assert 'Prediction mean:' in buf.getvalue()
        assert 'Prediction std:' in buf.getvalue()

    # Close the plot
    plt.close()



import torch
import torchvision
import os
from typing import Tuple
from torch import nn


def setup_model(config: dict) -> Tuple[nn.Module, int]:
    """Returns the model specified in the configuration file, with the last
    layer corresponding to the number of screens.

    Parameters
    ----------
    config : dict
        Dictionary containing the configuration

    Returns
    -------
    model : torch.nn.Module
        The model with the loaded state
    last_model_state : int
        The number of the last model state
    """

    model_name = config['model']['name']
    model_type = config['model']['type']
    pretrained = config['model']['pretrained']
    nscreens = config['speckle']['nscreens']
    data_directory = config['speckle']['datadirectory']

    if model_type in ['resnet18', 'resnet50', 'resnet152']:
        return get_a_resnet(nscreens, data_directory, model_name, model_type,
                            pretrained)
    else:
        raise ValueError(f'Unknown model {model_name}')


def get_a_resnet(nscreens: int, datadirectory: str, model_name: str,
                 model_type: str, pretrained: bool) -> Tuple[nn.Module, int]:
    """Returns a pretrained ResNet model, with the last layer corresponding to
    the number of screens.

    Parameters
    ----------
    nscreens : int
        Number of screens
    datadirectory : str
        Path to the directory containing the data
    model_name : str
        The name of the model
    model_type : str
        The type of the ResNet
    pretrained : bool
        Whether to use a pretrained model or not

    Returns
    -------
    model : torch.nn.Module
        The model with the loaded state
    last_model_state : int
        The number of the last model state
    """

    if model_type == 'resnet18':
        model = torchvision.models.resnet18(pretrained=pretrained)
    elif model_type == 'resnet50':
        model = torchvision.models.resnet50(pretrained=pretrained)
    elif model_type == 'resnet152':
        model = torchvision.models.resnet152(
            weights='IMAGENET1K_V2' if pretrained else None)
    else:
        raise ValueError(f'Unknown model {model_type}')

    # Give it its name
    model.name = model_name
    # put it in evaluation mode
    model.eval()

    # Change the model to process black and white input
    model.conv1 = torch.nn.Conv2d(1,
                                  64,
                                  kernel_size=(7, 7),
                                  stride=(2, 2),
                                  padding=(3, 3),
                                  bias=False)
    # Add a final layer to predict the output
    model.fc = torch.nn.Sequential(torch.nn.Linear(2048, nscreens),
                                   torch.nn.Sigmoid())

    return load_model_state(model, datadirectory)


def load_model_state(model: nn.Module,
                     datadirectory: str) -> Tuple[nn.Module, int]:
    """Loads the model state from the given directory.

    Parameters
    ----------
    model : torch.nn.Module
        The model to load the state into
    datadirectory : str
        The directory where the model states are stored

    Returns
    -------
    model : torch.nn.Module
        The model with the loaded state
    last_model_state : int
        The number of the last model state
    """

    # If no model is stored, create the folder
    if not os.path.isdir(f'{datadirectory}/{model.name}_states'):
        os.mkdir(f'{datadirectory}/{model.name}_states')

    # check what is the last model state
    try:
        last_model_state = sorted([
            int(file_name.split('.pth')[0].split('_')[-1])
            for file_name in os.listdir(f'{datadirectory}/{model.name}_states')
        ])[-1]
    except Exception as e:
        print(f'Warning: {e}')
        last_model_state = 0

    if last_model_state > 0:
        print(f'Loading model at epoch {last_model_state}')
        model.load_state_dict(
            torch.load(
                f'{datadirectory}/{model.name}_states/{model.name}_{last_model_state}.pth'
            ))
        return model, last_model_state
    else:
        print('No pretrained model to load')
        return model, 0


def setup_loss(config: dict) -> nn.Module:
    """Returns the criterion specified in the configuration file.

    Parameters
    ----------
    config : dict
        Dictionary containing the configuration

    Returns
    -------
    criterion : torch.nn.Module
        The criterion with the loaded state
    """

    criterion_name = config['hyppar']['loss']
    if criterion_name == 'BCELoss':
        return torch.nn.BCELoss()
    elif criterion_name == 'MSELoss':
        return torch.nn.MSELoss()
    else:
        raise ValueError(f'Unknown criterion {criterion_name}')


def setup_optimizer(config: dict, model: nn.Module) -> nn.Module:
    """Returns the optimizer specified in the configuration file.

    Parameters
    ----------
    config : dict
        Dictionary containing the configuration
    model : torch.nn.Module
        The model to optimize

    Returns
    -------
    optimizer : torch.nn.Module
        The optimizer with the loaded state
    """

    optimizer_name = config['hyppar']['optimizer']
    if optimizer_name == 'Adam':
        return torch.optim.Adam(model.parameters(), lr=config['hyppar']['lr'])
    elif optimizer_name == 'SGD':
        return torch.optim.SGD(model.parameters(), lr=config['hyppar']['lr'])
    else:
        raise ValueError(f'Unknown optimizer {optimizer_name}')



import torch
from speckcn2.mlops import score


def test_score():
    # Create a dummy model
    dmodel = torch.nn.Sequential(torch.nn.Flatten(),
                                 torch.nn.Linear(32 * 32, 1024),
                                 torch.nn.ReLU(), torch.nn.Linear(1024, 8))
    dmodel.name = 'test_model'
    # Create a dummy test loader
    dataset = [(torch.rand(1, 1, 32, 32), torch.rand(1, 8)) for _ in range(32)]
    test_loader = torch.utils.data.DataLoader(dataset)
    device = torch.device('cpu')
    criterion = torch.nn.MSELoss()

    def recover_tag(x):
        return x

    # Call the function
    test_tags = score(dmodel,
                      test_loader,
                      device,
                      criterion,
                      recover_tag,
                      data_dir='speckcn2/assets/test',
                      nimg_plot=1)

    # Assert the expected output
    assert len(test_tags) == 32
    assert test_tags[0].shape[0] == 8



import torch
import matplotlib.pyplot as plt
from PIL import Image
import os
import numpy as np
from typing import Callable, List, Tuple
import torchvision.transforms as transforms
from torch.utils.data import DataLoader


def assemble_transform(conf: dict) -> transforms.Compose:
    """Assembles the transformation to apply to each image.

    Parameters
    ----------
    conf : dict
        Dictionary containing the configuration

    Returns
    -------
    transform : torchvision.transforms.Compose
        Transformation to apply to the images
    """
    list_transforms = []

    if conf['preproc']['centercrop']:
        # Take only the center of the image
        list_transforms.append(
            transforms.CenterCrop(conf['preproc']['centercrop']))

    if conf['preproc']['polarize']:
        # make the image larger for better polar conversion
        list_transforms.append(transforms.Resize(conf['preproc']['polresize']))
        # Convert to polar coordinates
        list_transforms.append(PolarCoordinateTransform())

    if conf['preproc']['randomrotate'] and not conf['preproc']['polarize']:
        # Randomly rotate the image, since it is symmetric (not if it is polarized)
        list_transforms.append(transforms.RandomRotation(degrees=(-180, 180)))

    if conf['preproc']['resize']:
        # Optionally, downscale it
        list_transforms.append(
            transforms.Resize(
                (conf['preproc']['resize'], conf['preproc']['resize'])))

    if conf['preproc']['equivariant']:
        # Apply the equivariant transform
        list_transforms.append(ShiftRowsTransform())

    list_transforms.append(transforms.ToTensor())

    return transforms.Compose(list_transforms)


class PolarCoordinateTransform(torch.nn.Module):
    """Transform a Cartesian image to polar coordinates."""

    def __init__(self):
        super(PolarCoordinateTransform, self).__init__()

    def forward(self, img):
        """ forward method of the transform
        Args:
            img (PIL Image or Tensor): Image to be scaled.

        Returns:
            PIL Image or Tensor: Rescaled image.
        """

        img = np.array(img)  # Convert PIL image to NumPy array

        # Assuming img is a grayscale image
        height, width = img.shape
        polar_image = np.zeros_like(img)

        # Center of the image
        center_x, center_y = width // 2, height // 2

        # Maximum possible value of r
        max_r = np.sqrt(center_x**2 + center_y**2)

        # Create a grid of (x, y) coordinates
        y, x = np.ogrid[:height, :width]

        # Shift the grid so that the center of the image is at (0, 0)
        x = x - center_x
        y = y - center_y

        # Convert Cartesian to Polar coordinates
        r = np.sqrt(x**2 + y**2)
        theta = np.arctan2(y, x)

        # Rescale r to [0, height)
        r = np.round(r * (height - 1) / max_r).astype(int)

        # Rescale theta to [0, width)
        theta = np.round((theta + 2 * np.pi) % (2 * np.pi) * (width - 1) /
                         (2 * np.pi)).astype(int)

        # Use a 2D histogram to accumulate all values that map to the same polar coordinate
        histogram, _, _ = np.histogram2d(theta.flatten(),
                                         r.flatten(),
                                         bins=[height, width],
                                         range=[[0, height], [0, width]],
                                         weights=img.flatten())

        # Count how many Cartesian coordinates map to each polar coordinate
        counts, _, _ = np.histogram2d(theta.flatten(),
                                      r.flatten(),
                                      bins=[height, width],
                                      range=[[0, height], [0, width]])

        # Take the average of all values that map to the same polar coordinate
        polar_image = histogram / counts

        # Handle any divisions by zero
        polar_image[np.isnan(polar_image)] = 0

        # Crop the large r part that is not used
        for x in range(width - 1, -1, -1):
            # If the column contains at least one non-black pixel
            if np.any(polar_image[:, x] != 0):
                # Crop at this x position
                polar_image = polar_image[:, :x]
                break

        # reconvert to PIL image before returning
        return Image.fromarray(polar_image)


class ShiftRowsTransform(torch.nn.Module):
    """Shift the rows of an image such that the row with the smallest sum is at
    the bottom."""

    def __init__(self):
        super(ShiftRowsTransform, self).__init__()

    def forward(self, img):
        img = np.array(img)  # Convert PIL image to NumPy array

        # Find the row with the largest sum
        row_sums = np.sum(img, axis=1)
        max_sum_row_index = np.argmin(row_sums)

        # Shift all rows respecting periodicity
        shifted_img = np.roll(img, -max_sum_row_index, axis=0)

        # reconvert to PIL image before returning
        return Image.fromarray(shifted_img)


def prepare_data(conf: dict,
                 nimg_print: int = 5,
                 nreps: int = 1) -> Tuple[List, List]:
    """If not already available, preprocesses the data by loading images and
    tags from the given directory, applying a transformation to the images.

    Parameters
    ----------
    conf : dict
        Dictionary containing the configuration
    nimg_print: int
        Number of images to print
    nreps: int
        Number of repetitions for each image

    Returns
    -------
    all_images : list
        List of all images
    all_tags : list
        List of all tags
    """
    datadirectory = conf['speckle']['datadirectory']
    mname = conf['model']['name']

    # Dummy transformation to get the original image
    transform_orig = transforms.Compose([
        transforms.CenterCrop(conf['preproc']['centercrop']),
        transforms.ToTensor(),
    ])

    # and get the transformation to apply to each image
    transform = assemble_transform(conf)

    # First, check if the data has already been preprocessed
    if os.path.exists(os.path.join(datadirectory, f'all_images_{mname}.pt')):
        print('*** Loading preprocessed data')
        # If so, load it
        all_images = torch.load(
            os.path.join(datadirectory, f'all_images_{mname}.pt'))
        all_tags = torch.load(
            os.path.join(datadirectory, f'all_tags_{mname}.pt'))
        return all_images, all_tags

    # Otherwise, preprocess the raw data:
    file_list = [
        file_name for file_name in os.listdir(datadirectory)
        if '.txt' in file_name and 'tag' not in file_name
    ]
    # and randomly shuffle them
    np.random.shuffle(file_list)
    # I can also do data augmentation, by using each file multiple times (only if I am also doing a random rotation)
    file_list = file_list * nreps
    all_images = []
    all_tags = []
    if nimg_print > 0:
        show_image = True
        # and create a folder to store the images
        if not os.path.isdir(f'{datadirectory}/images_{mname}'):
            os.mkdir(f'{datadirectory}/images_{mname}')
    else:
        show_image = False

    # Load each text file as an image
    for counter, file_name in enumerate(file_list):
        # Construct the full path to the file
        file_path = os.path.join(datadirectory, file_name)

        # Open the text file as an image using PIL
        with open(file_path, 'r') as text_file:
            lines = text_file.readlines()

            pixel_values = np.array([[
                0 if value == 'NaN' or value == 'NaN\n' else float(value)
                for value in line.split(',')
            ] for line in lines],
                                    dtype=np.float32)

            # Create the image
            image_orig = Image.fromarray(pixel_values, mode='F')

            # Apply the transformation
            image = transform(image_orig)
            image_orig = transform_orig(image_orig)

            # and add it to the collection
            all_images.append(image)

            # Then load the screen tags
            #tagname = file_name.replace('.txt', '_tag.txt')
            base_name, _, _ = file_name.rpartition('_')
            tagname = base_name + '_tag.txt'

            if os.path.exists(os.path.join(datadirectory, tagname)):
                tags = np.loadtxt(os.path.join(datadirectory, tagname),
                                  delimiter=',',
                                  dtype=np.float32)

                # Plot the image using maplotlib
                if counter > nimg_print:
                    show_image = False
                if show_image:
                    fig, axs = plt.subplots(1, 3, figsize=(15, 5))
                    # Plot the original image
                    axs[0].imshow(image_orig.squeeze(), cmap='bone')
                    axs[0].set_title(f'Training Image {file_name}')
                    # Plot the preprocessd image
                    axs[1].imshow(image.squeeze(), cmap='bone')
                    axs[1].set_title('Processed as')
                    axs[1].set_xlabel(r'$r$')
                    axs[1].set_ylabel(r'$\theta$')

                    # Plot the tags
                    axs[2].plot(tags, 'o')
                    axs[2].set_yscale('log')
                    axs[2].set_title('Screen Tags')
                    axs[2].legend()

                    fig.subplots_adjust(wspace=0.3)
                    plt.savefig(
                        f'{datadirectory}/images_{mname}/{counter}.png')
                    plt.close()

                # Preprocess the tags
                tags = np.log10(tags)

                # Add the tag to the colleciton
                all_tags.append(tags)
            else:
                print(f'*** Warning: tag file {tagname} not found.')

    # Finally, store them before returning
    torch.save(all_images, os.path.join(datadirectory,
                                        f'all_images_{mname}.pt'))
    torch.save(all_tags, os.path.join(datadirectory, f'all_tags_{mname}.pt'))

    print('*** Preprocessing complete.', flush=True)

    return all_images, all_tags


def normalize_tags(
    all_images: List[torch.tensor], all_tags: List[np.ndarray]
) -> Tuple[List[Tuple[torch.tensor, np.ndarray]], Callable[
    [np.ndarray], np.ndarray], Callable[[np.ndarray], np.ndarray]]:
    """Normalize the tags to be between 0 and 1.

    Parameters
    ----------
    all_images : list
        List of all images
    all_tags : list
        List of all tags

    Returns
    -------
    dataset : list
        List of tuples (image, normalized_tag)
    normalize_tag : function
        Function to normalize a tag
    recover_tag : function
        Function to recover a tag
    """

    # I process the arrays of tags, such that each tag is a number between 0 and 1
    tags = [tag for tag in all_tags]
    #tags = np.array(tags, dtype=object)
    tags = np.stack(tags)

    # I need to find the minimum and maximum value of each tag
    # I will use this to normalize the tags
    min_tags = np.min(tags, axis=0)
    max_tags = np.max(tags, axis=0)
    range_tags = max_tags - min_tags

    def normalize_tag(tag):
        return (tag - min_tags) / range_tags

    def recover_tag(ntag):
        return ntag * range_tags + min_tags

    # And normalize the tags
    normalized_tags = np.array([normalize_tag(tag) for tag in tags])

    # I can now create the dataset
    dataset = [(image, tag) for image, tag in zip(all_images, normalized_tags)]

    return dataset, normalize_tag, recover_tag


def train_test_split(
        dataset: List[Tuple[torch.tensor, float]],
        batch_size: int = 32,
        train_test_split: float = 0.8) -> Tuple[DataLoader, DataLoader]:
    """Splits the data into training and testing sets.

    Parameters
    ----------
    dataset : list
        List of tuples (image, tag)
    batch_size : int
        Batch size for the data loaders
    train_test_split: float
        Fraction of the data to use for training

    Returns
    -------
    train_loader : torch.utils.data.DataLoader
        Training data loader
    test_loader : torch.utils.data.DataLoader
        Testing data loader
    """

    train_size = int(train_test_split * len(dataset))
    print(
        f'*** There are {len(dataset)} images in the dataset, {train_size} for training and {len(dataset)-train_size} for testing.'
    )
    train_loader = torch.utils.data.DataLoader(dataset[:train_size],
                                               batch_size=batch_size,
                                               shuffle=True,
                                               num_workers=2)
    test_loader = torch.utils.data.DataLoader(dataset[train_size:],
                                              batch_size=1,
                                              shuffle=False,
                                              num_workers=2)

    return train_loader, test_loader



import yaml
from speckcn2.io import load_config


def test_load_config():
    config_file_path = 'speckcn2/assets/test/test_config.yaml'
    with open(config_file_path, 'r') as file:
        expected_config = yaml.safe_load(file)

    actual_config = load_config(config_file_path)
    assert actual_config == expected_config



import os
import torch
import matplotlib.pyplot as plt
from torch import nn, optim, Tensor
from torch.utils.data import DataLoader
from typing import Callable, List, Tuple


def train(model: nn.Module, last_model_state: int, conf: dict,
          train_loader: DataLoader, device: torch.device,
          optimizer: optim.Optimizer,
          criterion: nn.Module) -> Tuple[nn.Module, float]:
    """Trains the model for the given number of epochs.

    Parameters
    ----------
    model : torch.nn.Module
        The model to train
    last_model_state : int
        The number of the last model state
    conf : dict
        Dictionary containing the configuration
    train_loader : torch.utils.data.DataLoader
        The training data loader
    device : torch.device
        The device to use
    optimizer : torch.optim
        The optimizer to use
    criterion : torch.nn
        The loss function to use

    Returns
    -------
    model : torch.nn.Module
        The trained model
    average_loss : float
        The average loss of the last epoch
    """

    final_epoch = conf['hyppar']['maxepochs']
    save_every = conf['model']['save_every']
    datadirectory = conf['speckle']['datadirectory']

    print(f'Training the model from epoch {last_model_state} to {final_epoch}')
    average_loss = 0.0
    for epoch in range(last_model_state, final_epoch):
        total_loss = 0.0
        for i, (inputs, tags) in enumerate(train_loader):
            # Move input and label tensors to the device
            inputs = inputs.to(device)
            tags = tags.to(device)

            # Zero out the optimizer
            optimizer.zero_grad()

            # Forward pass
            outputs = model(inputs)
            loss = criterion(outputs, tags)

            # Backward pass
            loss.backward()
            optimizer.step()

            # Accumulate the loss
            total_loss += loss.item()

        # Calculate average loss for the epoch
        average_loss = total_loss / len(train_loader)

        # Print the average loss for every epoch
        print(
            f'Epoch {epoch+1}/{final_epoch}, Average Loss: {average_loss:.5f}',
            flush=True)

        if (epoch + 1) % save_every == 0 or epoch == final_epoch - 1:
            # Save the model state
            torch.save(
                model.state_dict(),
                f'{datadirectory}/{model.name}_states/{model.name}_{epoch+1}.pth'
            )

    return model, average_loss


def score(model: nn.Module,
          test_loader: DataLoader,
          device: torch.device,
          criterion: nn.Module,
          recover_tag: Callable[[Tensor], Tensor],
          data_dir: str,
          nimg_plot: int = 1000) -> List[Tensor]:
    """Tests the model.

    Parameters
    ----------
    model : torch.nn.Module
        The model to test
    test_loader : torch.utils.data.DataLoader
        The testing data loader
    device : torch.device
        The device to use
    criterion : torch.nn
        The loss function to use
    recover_tag : function
        Function to recover a tag
    data_dir : str
        The directory where the data is stored
    nimg_plot : int
        Number of images to plot

    Returns
    -------
    test_tags : list
        List of all the predicted tags of the test set
    """

    with torch.no_grad():
        test_tags = []
        # create the directory where the images will be stored
        if not os.path.isdir(f'{data_dir}/score_{model.name}'):
            os.mkdir(f'{data_dir}/score_{model.name}')
        for idx, (inputs, tags) in enumerate(test_loader):
            # Move input and label tensors to the device
            inputs = inputs.to(device)
            tags = tags.to(device)

            # Forward pass
            outputs = model(inputs)
            loss = criterion(outputs, tags)

            # Print the loss for every epoch
            print(f'Item {idx} loss: {loss.item():.4f}')

            if idx < nimg_plot:
                # Plot the image and output side by side
                fig, axs = plt.subplots(1, 3, figsize=(9, 2.5))
                axs[0].imshow(inputs[0].detach().cpu().squeeze(), cmap='bone')
                axs[0].set_title('Loss: {:.4f}'.format(loss.item()))
                axs[1].plot(10**(recover_tag(tags[0].detach().cpu().numpy())),
                            'o',
                            label='True')
                axs[1].plot(10**(recover_tag(
                    outputs[0].detach().cpu().numpy())),
                            '.',
                            color='tab:red',
                            label='Predicted')
                axs[1].set_yscale('log')
                axs[1].set_title('Screen Tags')
                axs[1].legend()
                axs[2].plot(tags[0].detach().cpu().numpy(), 'o', label='True')
                axs[2].plot(outputs[0].detach().cpu().numpy(),
                            '.',
                            color='tab:red',
                            label='Predicted')
                axs[2].set_title('Unnormalized out')
                axs[2].set_ylim(0, 1)
                axs[2].legend()
                plt.savefig(f'{data_dir}/score_{model.name}/score_{idx}.png')
                plt.close()

            # and get all the tags for statistic analysis
            for tag in outputs:
                test_tags.append(tag)

    return test_tags



import os
import torch
import numpy as np
from speckcn2.preprocess import prepare_data


def test_prepare_data(tmpdir):
    # Define the test data directory
    test_data_dir = 'speckcn2/assets/test'

    conf = {
        'speckle': {
            'datadirectory': test_data_dir
        },
        'model': {
            'name': 'test_model'
        },
        'preproc': {
            'polarize': False,
            'equivariant': False,
            'randomrotate': False,
            'centercrop': 100,
            'resize': 100,
        }
    }

    # Call the function
    all_images, all_tags = prepare_data(conf)

    # Assert the expected output
    assert isinstance(all_images, list)
    assert isinstance(all_tags, list)

    # Check if the preprocessed data files are saved
    assert os.path.exists(
        os.path.join(test_data_dir, 'all_images_test_model.pt'))
    assert os.path.exists(os.path.join(test_data_dir,
                                       'all_tags_test_model.pt'))

    # Load the preprocessed data files
    loaded_images = torch.load(
        os.path.join(test_data_dir, 'all_images_test_model.pt'))
    loaded_tags = torch.load(
        os.path.join(test_data_dir, 'all_tags_test_model.pt'))

    # Assert the loaded data matches the original data
    assert len(loaded_images) == len(all_images)
    assert len(loaded_tags) == len(all_tags)
    for i in range(len(all_images)):
        assert torch.all(torch.eq(loaded_images[i], all_images[i]))
        assert np.array_equal(loaded_tags[i], all_tags[i])

    # Clean up the preprocessed data files
    os.remove(os.path.join(test_data_dir, 'all_images_test_model.pt'))
    os.remove(os.path.join(test_data_dir, 'all_tags_test_model.pt'))



import os
import torch
import pytest
from speckcn2.mlmodels import load_model_state


@pytest.fixture
def model_and_directory(tmpdir):
    model = torch.nn.Module()
    model.name = 'test_model'
    datadirectory = tmpdir.mkdir('test_data')
    return model, str(datadirectory)


def test_load_model_state_no_model_folder(model_and_directory):
    model, datadirectory = model_and_directory

    # Call the load_model_state function
    loaded_model, last_state = load_model_state(model, datadirectory)

    # Check if the model folder is created
    model_folder = os.path.join(datadirectory, 'test_model_states')
    assert os.path.isdir(model_folder)

    # Check if the loaded model is the same as the input model
    assert loaded_model is model

    # Check if the last model state is 0
    assert last_state == 0


def test_load_model_state_existing_model_states(model_and_directory):
    model, datadirectory = model_and_directory

    # Create some dummy model state files
    model_folder = os.path.join(datadirectory, 'test_model_states')
    os.mkdir(model_folder)
    torch.save(model.state_dict(),
               os.path.join(model_folder, 'test_model_1.pth'))
    torch.save(model.state_dict(),
               os.path.join(model_folder, 'test_model_2.pth'))

    # Call the load_model_state function
    loaded_model, last_state = load_model_state(model, datadirectory)

    # Check if the loaded model is the same as the input model
    assert loaded_model is model

    # Check if the last model state is the highest numbered state file
    assert last_state == 2


def test_load_model_state_no_model_states(model_and_directory):
    model, datadirectory = model_and_directory

    # Remove any existing model state files
    model_folder = os.path.join(datadirectory, 'test_model_states')
    os.mkdir(model_folder)

    # Call the load_model_state function
    loaded_model, last_state = load_model_state(model, datadirectory)

    # Check if the loaded model is the same as the input model
    assert loaded_model is model

    # Check if the last model state is 0
    assert last_state == 0



# This is the __init__.py file for the rkm package

# Import any modules or subpackages here
from .preprocess import prepare_data, normalize_tags, train_test_split
from .mlmodels import setup_model, get_a_resnet, setup_loss, setup_optimizer
from .mlops import train, score
from .postprocess import tags_distribution

__all__ = [
    'prepare_data', 'normalize_tags', 'train_test_split', 'setup_model',
    'get_a_resnet', 'setup_loss', 'setup_optimizer', 'train', 'score',
    'tags_distribution'
]



import yaml


def load_config(config_file_path: str) -> dict:
    """Load the configuration file.

    Parameters
    ----------
    config_file_path : str
        Path to the .yaml configuration file

    Returns
    -------
    config : dict
        Dictionary containing the configuration
    """
    with open(config_file_path, 'r') as file:
        config = yaml.safe_load(file)
    return config



import os
import matplotlib.pyplot as plt
import numpy as np
from typing import Callable, Optional
from torch import Tensor
from torch.utils.data import Dataset
from torch import device as Device


def tags_distribution(
        conf: dict,
        dataset: Dataset,
        test_tags: Tensor,
        device: Device,
        rescale: bool = False,
        recover_tag: Optional[Callable[[Tensor], Tensor]] = None) -> None:
    """Plots the distribution of the tags.

    Parameters
    ----------
    conf : dict
        Dictionary containing the configuration
    dataset : torch.utils.data.Dataset
        The dataset used for training
    test_tags : torch.Tensor
        The predicted tags for the test dataset
    device : torch.device
        The device to use
    data_directory : str
        The directory where the data is stored
    rescale : bool, optional
        Whether to rescale the tags using recover_tag() or leave them between 0 and 1
    recover_tag : callable, optional
        Function to recover a tag
    """

    data_directory = conf['speckle']['datadirectory']
    model_name = conf['model']['name']

    # create a folder to store the plots
    if not os.path.isdir(f'{data_directory}/{model_name}_plots'):
        os.mkdir(f'{data_directory}/{model_name}_plots')

    train_tags = np.array([tag for _, tag in dataset])
    predic_tags = np.array([n.cpu().numpy() for n in test_tags])
    print(f'Data shape: {train_tags.shape}')
    print(f'Prediction shape: {predic_tags.shape}')
    print(f'Train mean: {train_tags.mean()}')
    print(f'Train std: {train_tags.std()}')
    print(f'Prediction mean: {predic_tags.mean()}')
    print(f'Prediction std: {predic_tags.std()}')
    # plot the distribution of each tag element
    fig, axs = plt.subplots(2, 4, figsize=(20, 10))
    for i in range(8):
        if rescale and recover_tag is not None:
            axs[i // 4, i % 4].hist(
                recover_tag(predic_tags[:, i]),
                bins=20,
                color='tab:red',
                density=True,
                #histtype='step',
                alpha=0.5,
                label='Model precitions')
            axs[i // 4, i % 4].hist(
                recover_tag(train_tags[:, i]),
                bins=20,
                color='tab:blue',
                density=True,
                #histtype='step',
                alpha=0.5,
                label='Training data')
        else:
            axs[i // 4, i % 4].hist(
                predic_tags[:, i],
                bins=20,
                color='tab:red',
                density=True,
                #histtype='step',
                alpha=0.5,
                label='Model precitions')
            axs[i // 4, i % 4].hist(
                train_tags[:, i],
                bins=20,
                color='tab:blue',
                density=True,
                #histtype='step',
                alpha=0.5,
                label='Training data')
        axs[i // 4, i % 4].set_title(f'Tag {i}')
    axs[0, 1].legend()
    plt.savefig(f'{data_directory}/{model_name}_plots/tags_distribution.png')
    plt.close()
